
#include "io.h"
#include "io_grf.h"
#include "io_file.h"
#include "zlib_win32.h"
#include "../console.h"

#include <tchar.h>
#include <stdio.h>
#include <windows.h>

#ifdef _WIN32
	#include <windows.h>
	#include "zlib_win32.h"
	HINSTANCE zlib_dll;
	#define zlib_inflateInit(strm) zlib_inflateInit_((strm),ZLIB_VERSION, sizeof(z_stream))
	#define zlib_deflateInit(strm, level) zlib_deflateInit_((strm),(level),ZLIB_VERSION,sizeof(z_stream))
	int (WINAPI* zlib_inflateInit_) OF((z_streamp strm, const char *version, int stream_size));
	int (WINAPI* zlib_inflate) OF((z_streamp strm, int flush));
	int (WINAPI* zlib_inflateEnd) OF((z_streamp strm));
	int (WINAPI* zlib_deflateInit_) (z_streamp strm, int level, const char *version, int stream_size);
	int (WINAPI* zlib_deflate) (z_streamp strm, int flush);
	int (WINAPI* zlib_deflateEnd) (z_streamp strm);
	int strcasecmp(const char *s1, const char *s2) {
		while(*s1 || *s2) {
			if((*s1 | 0x20) != (*s2 | 0x20)) {
				return *s1 - *s2;
			}
			s1++; s2++;
		}
		return 0;
	}
#else
	#include <zlib.h>
	#define zlib_inflateInit inflateInit
	#define zlib_inflate     inflate
	#define zlib_inflateEnd  inflateEnd
	#define zlib_deflateInit deflateInit
	#define zlib_deflate     deflate
	#define zlib_deflateEnd  deflateEnd
#endif

void io_init(void) {
	if(!zlib_dll) {
		zlib_dll = LoadLibrary(_T("zlib.dll"));
		(FARPROC)zlib_inflateInit_ = GetProcAddress(zlib_dll,"inflateInit_");
		(FARPROC)zlib_inflate      = GetProcAddress(zlib_dll,"inflate");
		(FARPROC)zlib_inflateEnd   = GetProcAddress(zlib_dll,"inflateEnd");
		(FARPROC)zlib_deflateInit_ = GetProcAddress(zlib_dll,"deflateInit_");
		(FARPROC)zlib_deflate      = GetProcAddress(zlib_dll,"deflate");
		(FARPROC)zlib_deflateEnd   = GetProcAddress(zlib_dll,"deflateEnd");
		if(zlib_dll == NULL) {
			MessageBox(NULL,_T("Can't load zlib.dll"),_T("grfio.c"),MB_OK);
			exit(1);
		}
	}
}

void io_final(void) {
	FreeLibrary(zlib_dll);
	zlib_inflateInit_ = NULL;
	zlib_inflate      = NULL;
	zlib_inflateEnd   = NULL;
}

struct io_data* io_create(void) {
	return (struct io_data*)calloc(sizeof(struct io_data),1);
}

void io_delete(struct io_data* io) {
	if(io->read_grf) io_grf_close(io->read_grf);
	free(io);
}

struct io_fp* io_read_open(struct io_data* io,const wchar_t *file) {
	struct io_fp *fp = calloc(sizeof(struct io_fp),1);
	if(fp == NULL) return NULL;

	fp->io       = io;
	fp->buf_data = io_read_all(io,file,&fp->buf_len);
	fp->mode     = IO_FP_READ;
	fp->buf_max  = fp->buf_len;
	if(fp->buf_data == NULL || fp->buf_len == 0) {
		free(fp->buf_data);
		free(fp);
		return NULL;
	}
	return fp;
}

struct io_fp* io_write_open(struct io_data* io,const wchar_t *file) {
	struct io_fp *fp = calloc(sizeof(struct io_fp),1);
	if(fp == NULL) return NULL;

	wcscpy(fp->filename,file);
	fp->io       = io;
	fp->mode     = IO_FP_WRITE;
	fp->buf_len  = 0;
	fp->buf_max  = 32768;
	fp->buf_data = malloc(32768);
	if(fp->buf_data == NULL) {
		free(fp);
		return NULL;
	}
	return fp;
}

int io_read(struct io_fp *fp,void *buf,int len) {
	int size = min(len,fp->buf_max - fp->buf_pos);
	if(size <= 0) return 0;
	memcpy(buf,fp->buf_data + fp->buf_pos,size);
	fp->buf_pos += size;
	return size;
}

int io_write(struct io_fp *fp,void *buf,int len) {
	if(fp == NULL || fp->buf_data == NULL) {
		return 0;
	}
	while(fp->buf_pos + len > fp->buf_max) {
		fp->buf_max *= 2;
		fp->buf_data = realloc(fp->buf_data,fp->buf_max);
		if(fp->buf_data == NULL) {
			return 0;
		}
	}
	memcpy(fp->buf_data + fp->buf_pos,buf,len);
	fp->buf_pos += len;
	fp->buf_len  = max(fp->buf_len,fp->buf_pos);
	return len;
}

int io_seek(struct io_fp *fp,int pos) {
	fp->buf_pos = pos;
	return 1;
}

int io_tell(struct io_fp *fp) {
	return fp->buf_pos;
}

int io_close(struct io_fp *fp) {
	if(fp->mode == IO_FP_WRITE) {
		io_write_all(fp->io,fp->filename,fp->buf_data,fp->buf_len);
	}
	free(fp->buf_data);
	free(fp);
	return 1;
}

int io_printfWA(struct io_fp* fp,const wchar_t *format, ...) {
	wchar_t buf_w[1024];
	char    buf_a[1024];
	va_list va;
	int     len;
	
	va_start(va,format);
	vswprintf(buf_w,format,va);
	io_unicode_to_mbcs(fp->io,buf_a,buf_w);
	len = strlen(buf_a);
	io_write(fp,buf_a,len);
	return len;
}


char* io_read_all(struct io_data* io,const wchar_t* file,int *len) {
	if(io->read_grf) {
		return io_grf_reads(io->read_grf,file,len);
	} else {
		return io_file_readall(file,len);
	}
}

int io_write_all(struct io_data* io,const wchar_t* file,void *data,int len) {
	if(io->write_grf) {
		return 0;
	} else {
		return io_file_writeall(file,data,len);
	}
}

int io_size(struct io_data* io,const wchar_t* file) {
	if(io->read_grf) {
		return io_grf_size(io->read_grf,file);
	} else {
		return io_file_size(file);
	}
}

void io_copy(struct io_data* io,const wchar_t *from,const wchar_t *to) {
	int size;
	void *buf = io_read_all(io,from,&size);
	if(buf) {
		io_write_all(io,to,buf,size);
		free(buf);
	}
}

int io_read_grf(struct io_data* io,const wchar_t *file) {
	if(io->read_grf) {
		io_grf_close(io->read_grf);
	}
	io->read_grf = io_grf_open(io,file);
	return 0;
}

int io_find_match(const wchar_t* expr,const wchar_t *file) {
	// console_outputf("match: expr:%s file:%s\n",expr,file);
	if(*file == 0) {
		return (*expr == 0);
	}
	while(expr[0] && file[0]) {
		if(expr[0] == '?') {
			expr++; file++;
		} else if(expr[0] == '*') {
			const wchar_t *f2 = file;
			if(*(++expr) == 0) {
				return 1;
			}
			do {
				if(io_find_match(expr,f2)) return 1;
			} while(*(++f2) != 0);
		} else if(expr[0] != file[0]) {
			return 0;
		} else {
			expr++; file++;
		}
	}
	if(expr[0] == 0 && file[0] == 0) {
		return 1;
	} else {
		return 0;
	}
}

struct io_find *io_find_start(struct io_data* io,const wchar_t *expr) {
	struct io_find *find = (struct io_find*)calloc(sizeof(struct io_find),1);
	wcsncpy(find->expr,expr,sizeof(find->expr));
	find->io = io;
	if(!wcschr(find->expr,'*') && !wcschr(find->expr,'?')) {
		// ChJ[hȂꍇ
		find->end = 1;
	}
	return find;
}

const wchar_t *io_find_next(struct io_find *find) {
	const wchar_t* ret;
	if(find->end) {
		if(find->end++ == 1) {
			return find->expr;
		} else {
			return NULL;
		}
	}
	while(1) {
		if(find->io->read_grf) {
			ret = io_grf_find(find);
		} else {
			ret = io_file_find(find);
		}
		if(ret == NULL) {
			find->end = 2;
			return NULL;
		} else if(io_find_match(find->expr,ret)) {
			return ret;
		}
	}
}

void io_find_end(struct io_find *find) {
	free(find);
}

wchar_t* io_mbcs_to_unicode(struct io_data* io,wchar_t *buf1,const char *buf2) {
	// Xbh̕R[hύX
	if(SetThreadLocale(MAKELCID(MAKELANGID(LANG_KOREAN,SUBLANG_KOREAN),SORT_DEFAULT)) == 0) {
		console_outputf("o_conv_unicode: failed to change locate\n");
	}
	MultiByteToWideChar(CP_THREAD_ACP,0,buf2,-1,buf1,1024);
	// Xbh̕R[hɖ߂
	SetThreadLocale(LOCALE_USER_DEFAULT);
	return buf1;
}

char* io_unicode_to_mbcs(struct io_data* io,char *buf1,const wchar_t *buf2) {
	// Xbh̕R[hύX
	if(SetThreadLocale(MAKELCID(MAKELANGID(LANG_KOREAN,SUBLANG_KOREAN),SORT_DEFAULT)) == 0) {
		console_outputf("o_conv_unicode: failed to change locate\n");
	}
	WideCharToMultiByte(CP_THREAD_ACP,0,buf2,-1,buf1,1024,NULL,NULL);
	// Xbh̕R[hɖ߂
	SetThreadLocale(LOCALE_USER_DEFAULT);
	return buf1;
}

int io_zlib_decode(char *dest, unsigned long* destLen, const char* source, unsigned long sourceLen)
{
	z_stream stream;
	int err;

	stream.next_in = (Bytef*)source;
	stream.avail_in = (uInt)sourceLen;
	/* Check for source > 64K on 16-bit machine: */
	if ((uLong)stream.avail_in != sourceLen) return Z_BUF_ERROR;

	stream.next_out = dest;
	stream.avail_out = (uInt)*destLen;
	if ((uLong)stream.avail_out != *destLen) return Z_BUF_ERROR;

	stream.zalloc = (alloc_func)0;
	stream.zfree = (free_func)0;

	err = zlib_inflateInit(&stream);
	if (err != Z_OK) return err;

	err = zlib_inflate(&stream, Z_FINISH);
	if (err != Z_STREAM_END) {
		zlib_inflateEnd(&stream);
		return err == Z_OK ? Z_BUF_ERROR : err;
	}
	*destLen = stream.total_out;

	err = zlib_inflateEnd(&stream);
	return err;
}

int io_zlib_encode(char *dest, unsigned long* destLen, const char* source, unsigned long sourceLen) {
	z_stream stream;
	int err;

	stream.next_in = (Bytef*)source;
	stream.avail_in = (uInt)sourceLen;
	/* Check for source > 64K on 16-bit machine: */
	if ((uLong)stream.avail_in != sourceLen) return Z_BUF_ERROR;

	stream.next_out = dest;
	stream.avail_out = (uInt)*destLen;
	if ((uLong)stream.avail_out != *destLen) return Z_BUF_ERROR;

	stream.zalloc = (alloc_func)0;
	stream.zfree = (free_func)0;

	err = zlib_deflateInit(&stream,Z_DEFAULT_COMPRESSION);
	if (err != Z_OK) return err;

	err = zlib_deflate(&stream, Z_FINISH);
	if (err != Z_STREAM_END) {
		zlib_inflateEnd(&stream);
		return err == Z_OK ? Z_BUF_ERROR : err;
	}
	*destLen = stream.total_out;

	err = zlib_deflateEnd(&stream);
	return err;
}


