#include <stdio.h>

#include "vram.h"
#include "io/io.h"

#include <vfw.h>
#pragma comment(lib, "vfw32.lib")

struct vram_data {
	PIXEL* data;
	int size_x;
	int size_y;
	struct io_fp *fp;
	int depth;
	HDC hdc;
	HBITMAP hbitmap;
	BITMAPINFO bmpinfo;
	unsigned char *line_buf;
	struct avi_data {
		int current_frame;
		PAVIFILE pavi;
		PAVISTREAM pstm;
		BITMAPINFOHEADER format;
		unsigned char *write_buf;
	} avi;
};

HDC vram_gethdc(struct vram_data *vram) {
	return vram->hdc;
}

struct vram_data* vram_alloc(int x,int y) {
	struct vram_data* vram;
	if(x<=0 || y<=0) { return NULL; }

	vram = (struct vram_data *)calloc(sizeof(struct vram_data),1);
	if(vram == NULL) {
		return NULL;
	}
	vram->size_x = x;
	vram->size_y = y;

#ifdef _WIN32
	vram_bitmapinfo(&vram->bmpinfo.bmiHeader,x,-y,sizeof(PIXEL)*8);
	vram->hdc     = CreateCompatibleDC(NULL);
	vram->hbitmap = CreateDIBSection(vram->hdc,&vram->bmpinfo,DIB_RGB_COLORS,(void **)(&vram->data),NULL,0);
	SelectObject(vram->hdc,vram->hbitmap);
#else
	vram->data = malloc(x * y * sizeof(PIXEL));
	if(vram->data == NULL) {
		free(vram->data);
		return NULL;
	}
#endif
	vram->line_buf = malloc(4 * x);
	return vram;
}

void vram_free(struct vram_data* vram) {
	if(vram != NULL) {
#ifdef _WIN32
		DeleteObject(vram->data);
		DeleteDC(vram->hdc);
#else
		free(vram->data);
#endif
		free(vram->line_buf);
		if(vram->fp != NULL) {
			io_close(vram->fp);
		}

		free(vram);
	}
}

void vram_setpixel(struct vram_data* vram,int x,int y,PIXEL color) {
	if(vram == NULL) { return; }
	
	vram->data[y * vram->size_x + x] = color;
}

PIXEL vram_getpixel(struct vram_data* vram,int x,int y) {
	if(vram == NULL) { return 0; }
	
	return vram->data[y * vram->size_x + x];
}

struct vram_size* vram_getsize(struct vram_data* vram,struct vram_size* size) {
	if(vram == NULL) {
		size->size_x = 0;
		size->size_y = 0;
	} else {
		size->size_x = vram->size_x;
		size->size_y = vram->size_y;
	}
	return size;
}

void vram_bitmaphead(BITMAPFILEHEADER *head,int x,int y,int depth) {
	int line_size = ((depth / 8) * x + 3) & (~3);
	memcpy(&head->bfType,"BM",2);
	head->bfReserved1	= 0;
	head->bfReserved2	= 0;
	head->bfOffBits		= sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	head->bfSize		= head->bfOffBits + line_size * y;
}

void vram_bitmapinfo(BITMAPINFOHEADER *head,int x,int y,int depth) {
	int line_size = ((depth / 8) * x + 3) & (~3);
	head->biSize			= sizeof(BITMAPINFOHEADER);
	head->biWidth			= x;
	head->biHeight			= y;
	head->biPlanes			= 1;
	head->biBitCount		= depth;
	head->biCompression		= BI_RGB;
	head->biSizeImage		= line_size * y;
	head->biXPelsPerMeter	= 0; 
	head->biYPelsPerMeter	= 0;
	head->biClrUsed			= 0;
	head->biClrImportant	= 0;
}

struct vram_data* vram_bmpload_file(struct io_data *io,wchar_t *file) {
	unsigned char *buf;
	struct vram_data* vram;
	if(file == NULL) { return NULL; }
	if(!file[0])     { return NULL; }

	buf  = io_read_all(io,file,NULL);
	vram = vram_bmpload_mem(buf);
	free(buf);
	return vram;
}

struct vram_data* vram_bmpload_mem(unsigned char *buf) {
	int x,y;
	BITMAPINFOHEADER *bmp_info;
	struct vram_data* vram;
	int line_size;
	unsigned char *line_p;
	PIXEL *vram_p;

	if(buf == NULL) { return NULL; }
	if(buf[0] != 'B' || buf[1] != 'M') {
		return NULL;
	}
	buf += sizeof(BITMAPFILEHEADER);
	
	bmp_info = (BITMAPINFOHEADER*)buf;
	buf += bmp_info->biSize;

	if(bmp_info->biCompression != BI_RGB) {
		return NULL;
	}
	vram = vram_alloc(bmp_info->biWidth,bmp_info->biHeight);
	if(vram == NULL) {
		return NULL;
	}

	// dword Eɐ؂グ
	line_size = (bmp_info->biWidth*bmp_info->biBitCount/8 + 3) & ~3;
	vram_p = &vram->data[(vram->size_y-1)*vram->size_x];

	if(bmp_info->biBitCount == 8) {
		//  8Bit BMP
		int color_table_count;
		PIXEL color_table[256];
		if(bmp_info->biClrUsed == 0) {
			color_table_count = 256;
		} else {
			color_table_count = bmp_info->biClrUsed;
		}
		for(x=0;x<color_table_count;x++) {
			color_table[x]=vram_RGB(buf[2],buf[1],buf[0]);
			buf += 4;
		}
		for(y=0;y<bmp_info->biHeight;y++) {
			line_p = buf;
			for(x=0;x<bmp_info->biWidth;x++) {
				vram_p[x] = color_table[line_p[0]];
				line_p++;
			}
			buf += line_size;
			vram_p -= vram->size_x;
		}
		return vram;
	} else if(bmp_info->biBitCount == 24) {
		// 24Bit BMP
		for(y=0;y<bmp_info->biHeight;y++) {
			line_p = buf;
			for(x=0;x<bmp_info->biWidth;x++) {
				vram_p[x]=vram_RGB(line_p[2],line_p[1],line_p[0]);
				line_p += 3;
			}
			buf += line_size;
			vram_p -= vram->size_x;
		}
		return vram;
	} else {
		vram_free(vram);
		return NULL;
	}
}

struct vram_data *vram_load_color_table(unsigned char *data,int size_x,int size_y,PIXEL *ctbl) {
	struct vram_data* vram = vram_alloc(size_x,size_y);
	PIXEL *vram_pos;
	int x,y;
	if(vram == NULL) { return NULL; }
	for(y = 0;y < size_y ; y++) {
		vram_pos = &vram->data[vram->size_x * y];
		for(x = 0;x < size_x ; x++) {
			*(vram_pos++) = ctbl[*(data++)];
		}
	}
	return vram;
}

int vram_bmpsave_start(struct vram_data* vram,struct io_data* io,wchar_t *filename,int sizey,int depth) {
	BITMAPFILEHEADER bmp_header;
	BITMAPINFOHEADER bmp_info;
	vram->fp = io_write_open(io,filename);
	if(vram->fp == NULL) { return 0; }
	vram_bitmaphead(&bmp_header,vram->size_x,sizey,depth);
	io_write(vram->fp,&bmp_header,sizeof(BITMAPFILEHEADER));

	vram_bitmapinfo(&bmp_info,vram->size_x,sizey,depth);
	io_write(vram->fp,&bmp_info,sizeof(BITMAPINFOHEADER));

	vram->depth = depth;
	return 1;
}

// Ps̃f[^ϊ
int vram_bmpsave_line(struct vram_data *vram,PIXEL *pixel,unsigned char *buf) { 
	int x;
	int line_size;
	if(vram->depth == 16) {
		line_size = (2 * vram->size_x + 3) & (~3);
		for(x=0;x<vram->size_x;x++) {
			PIXEL vram_p2 = pixel[x];
			*(unsigned short*)buf = (unsigned short)(
				((vram_R(vram_p2)>>3)<<10) | ((vram_G(vram_p2)>>3)<<5) | (vram_B(vram_p2)>>3)
			);
			buf += sizeof(unsigned short);
		}
	} else if(vram->depth == 24) {
		line_size = (3 * vram->size_x + 3) & (~3);
		for(x=0;x<vram->size_x;x++) {
			PIXEL vram_p2 = pixel[x];
			buf[0] = (unsigned char)vram_B(vram_p2);
			buf[1] = (unsigned char)vram_G(vram_p2);
			buf[2] = (unsigned char)vram_R(vram_p2);
			buf += 3;
		}
		buf[0] = buf[1] = buf[2] = 0;
	} else if(vram->depth == 32) {
		line_size = 4 * vram->size_x;
		memcpy(buf,pixel,line_size);
	}
	return line_size;
}

int vram_bmpsave_data(struct vram_data* vram,int sizey) {
	int y;
	PIXEL *vram_p;
	if(vram->fp == NULL) { return 0; }

	vram_p = &vram->data[(sizey-1)*vram->size_x];
	for(y=0;y<sizey;y++) {
		int line_size = vram_bmpsave_line(vram,vram_p,vram->line_buf);
		io_write(vram->fp,vram->line_buf,line_size);
		vram_p -= vram->size_x;
	}
	return 1;
}

int vram_bmpsave_end(struct vram_data* vram) {
	if(vram->fp == NULL) { return 0; }
	io_close(vram->fp);
	vram->fp = NULL;
	return 1;
}

int vram_bmpsave(struct vram_data* vram,struct io_data* io,wchar_t *filename,int depth) {
	vram_bmpsave_start(vram,io,filename,vram->size_y,depth);
	vram_bmpsave_data(vram,vram->size_y);
	return vram_bmpsave_end(vram);
}

void vram_swap(int *a,int *b) {
	int c; c = *a; *a = *b; *b = c;
}

int vram_copy(struct vram_data* from,struct vram_data* to,RECT *from_r,RECT *to_r,int angle,int alpha) {
	int from_sizex,from_sizey,to_sizex,to_sizey;
	if(from == NULL || to == NULL) {
		return 0;
	}
	// 180 x] = x ] , y ]
	if(angle == 180 || angle == 270) {
		vram_swap((int*)&from_r->left,(int*)&from_r->right);
		vram_swap((int*)&from_r->top,(int*)&from_r->bottom);
		angle -= 180;
	}
	// to.left < to.right , to.top < to.bottom ɂȂ悤ɓւ
	if(to_r->top > to_r->bottom) {
		vram_swap((int*)&to_r->top,(int*)&to_r->bottom);
		vram_swap((int*)&from_r->top,(int*)&from_r->bottom);
	}
	if(to_r->left > to_r->right) {
		vram_swap((int*)&to_r->left,(int*)&to_r->right);
		vram_swap((int*)&from_r->left,(int*)&from_r->right);
	}
	// from , to ̑傫߂
	if(from_r->top < from_r->bottom) {
		from_sizey = from_r->bottom - from_r->top + 1;
	} else {
		from_sizey = from_r->top - from_r->bottom + 1;
	}
	if(from_r->left < from_r->right) {
		from_sizex = from_r->right - from_r->left + 1;
	} else {
		from_sizex = from_r->left - from_r->right + 1;
	}
	to_sizey = to_r->bottom - to_r->top + 1;
	to_sizex = to_r->right - to_r->left + 1;
	if(angle == 0) {
		int x,y;
		double pos_y=from_r->top;
		double pos_x;
		double ratio_x = (double)from_sizex  / to_sizex;
		double ratio_y = (double)from_sizey  / to_sizey;
		PIXEL *vram_1; 
		PIXEL *vram_2 = &to->data[to_r->top * to->size_x + to_r->left];
		if(from_r->top > from_r->bottom) {
			ratio_y = - ratio_y;
		}
		if(from_r->left > from_r->right) {
			ratio_x = - ratio_x;
		}
		if(alpha == 0) {
			;
		} else if(alpha == 255) {
			for(y = 0 ; y < to_sizey ; y++) {
				pos_x = 0;
				vram_1=&from->data[(int)(pos_y)*from->size_x + from_r->left];
				for(x = 0 ; x < to_sizex ; x++) {
					if(vram_A(vram_1[(int)(pos_x)]) < 128) {
						vram_2[x]=vram_1[(int)(pos_x)];
					}
					pos_x += ratio_x;
				}
				pos_y += ratio_y;
				vram_2 += to->size_x;
			}
		} else if(alpha < 255) {
			// At@uhL
			for(y = 0 ; y < to_sizey ; y++) {
				pos_x = 0;
				vram_1=&from->data[(int)(pos_y)*from->size_x + from_r->left];
				for(x = 0 ; x < to_sizex ; x++) {
					register PIXEL v1 , v2 , v3;
					v1 = vram_1[(int)(pos_x)];
					v2 = vram_2[x];
					if(vram_A(v1) < 128) {
						v3 = v2 & 0x00FF00FF;
						vram_2[x] =  (((alpha * ((v1&0x00FF00FF)-v3)) >> 8) + v3) & 0x00FF00FF;
						v3 = v2 & 0x0000FF00;
						vram_2[x] |= (((alpha * ((v1&0x0000FF00)-v3)) >> 8) + v3) & 0x0000FF00;
					}
					pos_x += ratio_x;
				}
				pos_y += ratio_y;
				vram_2 += to->size_x;
			}
		}
	} else if(angle == 90) {
		double ratio_x = (double)from_sizex / to_sizey;
		double ratio_y = (double)from_sizey / to_sizex;
		PIXEL *vram_1;
		PIXEL *vram_2 = &to->data[to_r->top * to->size_x + to_r->left];
		PIXEL *vram_2p;
		int y,x;
		double pos_x,pos_y;
		if(from_r->top > from_r->bottom) {
			ratio_y = - ratio_y;
		}
		if(from_r->left > from_r->right) {
			ratio_x = - ratio_x;
		}
		pos_y = from_r->top;
		if(alpha == 0) {
			;
		} else if(alpha == 255) {
			for(x = 0 ; x < to_sizex ; x++) { 
				vram_1  = &from->data[(int)pos_y * from->size_x + from_r->left];
				vram_2p = &vram_2[to->size_x * (to_sizey - 1)];
				pos_x = 0;
				for(y = 0 ; y < to_sizey ; y++) { 
					if(vram_A(vram_1[(int)(pos_x)]) < 128) {
						*vram_2p = vram_1[(int)pos_x];
					}
					pos_x += ratio_x;
					vram_2p -= to->size_x;
				}
				vram_2++;
				pos_y += ratio_y;
			}
		} else if(alpha < 255) {
			// At@uhL(At@lfrom)
			for(x = 0 ; x < to_sizex ; x++) { 
				vram_1  = &from->data[(int)pos_y * from->size_x + from_r->left];
				vram_2p = &vram_2[to->size_x * (to_sizey - 1)];
				pos_x = 0;
				for(y = 0 ; y < to_sizey ; y++) {
					register PIXEL v1 , v2 , v3;

					v1 = vram_1[(int)pos_x];
					v2 = *vram_2p;

					if(vram_A(v1) < 128) {
						// uh(32bit only)
						v3 = v2 & 0x00FF00FF;
						*vram_2p =  (((alpha * ((v1&0x00FF00FF)-v3)) >> 8) + v3) & 0x00FF00FF;
						v3 = v2 & 0x0000FF00;
						*vram_2p |= (((alpha * ((v1&0x0000FF00)-v3)) >> 8) + v3) & 0x0000FF00;
					}

					pos_x += ratio_x;
					vram_2p -= to->size_x;
				}
				vram_2++;
				pos_y += ratio_y;
			}
		}
	}
	return 1;
}

struct vram_data* vram_resample(struct vram_data* from) {
	struct vram_data *to;
	int   x,y,sx,sy,skip;
	PIXEL *vram_1;
	PIXEL *vram_2;
	PIXEL *vram_3;

	if(from == NULL) { return NULL; }
	sx = from->size_x / 2;
	sy = from->size_y / 2;
	to = vram_alloc(sx,sy);
	if(to == NULL)   { return NULL; }
	vram_1 = from->data;
	vram_2 = &from->data[from->size_x];
	vram_3 = to->data;
	skip   = (from->size_x & 1);
	
	for(y = sy ; y ; y--) {
		for(x = sx; x ; x--) {
			register PIXEL vram_4;
			vram_4  = (vram_1[0] & 0x00FF00FF) + (vram_1[1] & 0x00FF00FF);
			vram_4 += (vram_2[0] & 0x00FF00FF) + (vram_2[1] & 0x00FF00FF);
			*vram_3 = ((vram_4 + 0x00020002) >> 2) & 0x00FF00FF;

			vram_4   = (vram_1[0] & 0xFF00FF00) + (vram_1[1] & 0xFF00FF00);
			vram_4  += (vram_2[0] & 0xFF00FF00) + (vram_2[1] & 0xFF00FF00);
			*vram_3 |= ((vram_4 + 0x02000200) >> 2) & 0xFF00FF00;
			vram_1 += 2; vram_2 += 2; vram_3++;
		}
		if(skip) {
			vram_1++; vram_2++;
		}
		vram_1 += from->size_x;
		vram_2 += from->size_x;
	}
	return to;
}

void vram_drawtext(struct vram_data* vram,char *data,int x,int y,PIXEL color) {
	RECT rect;
	SetTextColor(vram->hdc,RGB(vram_R(color),vram_G(color),vram_B(color)));
	rect.left	= x;
	rect.top	= y;
	rect.bottom	= vram->size_y;
	rect.right	= vram->size_x;
	DrawTextA(vram->hdc,data,strlen(data),&rect,DT_NOCLIP);
}

void vram_square(struct vram_data* vram,int x,int y,int xe,int ye,PIXEL color) {
	int i,j;
	PIXEL *vram_p;
	if(vram == NULL) return;
	if(x < 0) { x = 0; }
	if(y < 0) { y = 0; }
	if(xe >= vram->size_x) { xe = vram->size_x - 1; }
	if(ye >= vram->size_y) { ye = vram->size_y - 1; }
	vram_p = &vram->data[y * vram->size_x];
	for(i = y;i <= ye ; i++) {
		for(j = x; j <= xe ; j++) {
			vram_p[j] = color;
		}
		vram_p += vram->size_x;
	}
}

#ifdef _WIN32

void vram_aviwrite_start(struct vram_data *vram,struct io_data* io,wchar_t *filename,int interval,int depth) {
	AVISTREAMINFO si;
	BITMAPINFOHEADER head;
	struct avi_data *avi = &vram->avi;
	AVIFileInit();

	io_write_all(io,filename," ",1);
	if (AVIFileOpenW(&avi->pavi,filename,OF_CREATE,NULL)) {
		MessageBoxA(NULL, "Error: AVIFileOpen", "G[" , MB_OK);
		return;
	}

	memset(&si,0,sizeof(AVISTREAMINFO));
	si.fccType        = streamtypeVIDEO;
	si.fccHandler     = 0;
	si.dwScale        = interval;
	si.dwRate         = 1000;
	si.rcFrame.right  = vram->size_x;
	si.rcFrame.bottom = vram->size_y;

	if(AVIFileCreateStream(avi->pavi,&avi->pstm,&si) != AVIERR_OK) {
		MessageBoxA(NULL, "Error: AVIFileCreateStream", "G[" , MB_OK);
		return;
	}
	vram_bitmapinfo(&head,vram->size_x,vram->size_y,depth);

	if(AVIStreamSetFormat(avi->pstm,0,&head,sizeof(BITMAPINFOHEADER))) {
		MessageBoxA(NULL, "Error: AVIStreamSetFormat", "G[" , MB_OK);
		return;
	}
	avi->current_frame = 0;
	avi->write_buf = (unsigned char*)malloc(head.biSizeImage + 4);
	vram->depth = depth;
}

int vram_aviwrite_frame(struct vram_data *vram) {
	struct avi_data *avi = &vram->avi;
	int y;
	unsigned char *p = vram->avi.write_buf;
	for(y = vram->size_y - 1; y >= 0 ; y--) {
		p += vram_bmpsave_line(vram,&vram->data[y * vram->size_x],p);
	}
	if(AVIStreamWrite(
		avi->pstm,avi->current_frame++,1,vram->avi.write_buf,
		p - vram->avi.write_buf,AVIIF_KEYFRAME,NULL,NULL
	) != AVIERR_OK) {
		MessageBoxA(NULL, "Error: AVIStreamWrite", "G[" , MB_OK);
		// AviWriteEnd();
		return 0;
	} else {
		// console_outputf("Ok Freme %d : %d byte\n",avi->current_frame - 1,p - vram->avi.write_buf);
		return 1;
	}
}

void vram_aviwrite_close(struct vram_data *vram) {
	struct avi_data *avi = &vram->avi;
	free(avi->write_buf);
	AVIStreamRelease(avi->pstm);
	AVIFileRelease(avi->pavi);
	AVIFileExit();
}

#endif
